unit Customer;

interface

uses
  BUGLists,
  System.Collections,
  System.ComponentModel;

type

{$REGION 'TCustomer class'}
  TCustomer = class( TObject, ICustomer, IComparable, IEditableObject )
  strict private

    {$REGION 'Instance field variables'}
      FBackupData: TCustomer;
      FEditing: boolean;
      _owner: IList;
      _newCustomer: boolean;
    {$ENDREGION}

    {$REGION 'Property fields'}
    FTitle: &string;
    FFirstName: &string;
    FLastName: &string;
    {$ENDREGION}

    {$REGION 'IComparable members'}
    function CompareTo( p_customer: TCustomer ): integer; overload;
    function CompareTo( p_object: TObject ): integer; overload;
    {$ENDREGION}

    {$REGION 'Utility functions'}
    function IsChanged: boolean;
    {$ENDREGION}
  public

    {$REGION 'Default string representation'}
    function ToString: string; override;
    {$ENDREGION}

    {$REGION 'IEditableObject members'}
    procedure BeginEdit;
    procedure CancelEdit;
    procedure EndEdit;
    {$ENDREGION}

  published

    {$REGION 'property getters & setters'}
    function get_FullName: string;
    procedure set_FirstName(Value: string);
    procedure set_LastName(Value: string);
    procedure set_Title(Value: string);
    function get_FirstName: string;
    function get_LastName: string;
    function get_Title: string;
    {$ENDREGION}

  published

    constructor Create(p_owner: IList);

    {$REGION 'ICustomer members'}
    property FirstName: string read get_FirstName write set_FirstName;
    property LastName: string read get_LastName write set_LastName;
    property Title: string read get_Title write set_Title;
    property FullName: string read get_FullName;
    {$ENDREGION}

  end;
{$ENDREGION}

{$REGION 'Comparers' }

  TCompareCustomerByTitle = class( TObject, IComparer )
  protected
    function Compare( l_customer1, l_customer2: TObject ): integer;
  end;

  TCompareCustomerByFirstName = class( TObject, IComparer )
  protected
    function Compare( l_customer1, l_customer2: TObject ): integer;
  end;

  TCompareCustomerByLastName = class( TObject, IComparer )
  protected
    function Compare( p_customer1, p_customer2: TObject ): integer;
  end;

  TCompareCustomer = class( TObject, IComparer )
  protected
    function Compare( p_customer1, p_customer2: TObject ): integer;
  end;
{$ENDREGION}

implementation

uses
  CustomerList;

{$REGION 'TCustomer class'}

{$REGION 'IEditableObject members'}
procedure TCustomer.BeginEdit;
begin
  if ( not FEditing ) then
  begin
    if ( not Assigned( FBackupData ) ) then
    begin
      FBackupData := TCustomer.Create( nil );
    end;
    FBackupData.FTitle := FTitle;
    FBackupData.FFirstName := FFirstName;
    FBackupData.FLastName := FLastName;
    FEditing := True;
  end;
end;

procedure TCustomer.CancelEdit;
begin
  if ( FEditing ) then
  begin
    if ( Assigned( FBackupData ) ) then
    begin
      FTitle := FBackupData.FTitle;
      FFirstName := FBackupData.FFirstName;
      FLastName := FBackupData.FLastName;
    end;
    if ( _newCustomer ) then
    begin
      _owner.Remove( Self );
    end;
    FEditing := False;
  end;
end;

procedure TCustomer.EndEdit;
var
  l_position: integer;
begin
  _newCustomer := False;
  FEditing := False;
  if ( Self.IsChanged ) then
  begin
    // Update controls
    l_position := _owner.IndexOf( Self );
    _owner[ l_position ] := Self;
  end;
end;
{$ENDREGION}

{$REGION 'IComparable members'}
function TCustomer.CompareTo( p_object: TObject ): integer;
begin
  if ( p_object is TCustomer ) then
  begin
    Result := Self.CompareTo( TCustomer( p_object ) );
  end
  else
  begin
    raise ArgumentException.Create( 'object is not a Customer' );
  end;
//  Result := FullName.CompareTo( TCustomer( p_object ).FullName );
end;

{function TCustomer.CompareTo( p_object: TObject ): integer;
begin
  Result := TCustomerList( _owner ).CompareCustomers( Self, TCustomer( p_object ) );
end;}
{$ENDREGION}

{$REGION 'Alternative CompareTo method'}
function TCustomer.CompareTo(p_customer: TCustomer): integer;
var
  l_mySortName, l_objectsSortName: string;
begin
  l_mySortName := LastName + FirstName + Title;
  l_objectsSortName := TCustomer( p_customer ).LastName +
                       TCustomer( p_customer ).FirstName +
                       TCustomer( p_customer ).Title;
  Result := l_mySortName.CompareTo( l_objectsSortName );
end;
{$ENDREGION}

{$REGION 'Constructor'}
constructor TCustomer.Create(p_owner: IList);
begin
  inherited Create;
  if ( Assigned( p_owner ) ) then
  begin
    _owner := p_owner;
  end;
  _newCustomer := True;
end;
{$ENDREGION}

{$REGION 'ICustomer members'}
function TCustomer.get_FirstName: string;
begin
  Result := Self.FFirstName;
end;

function TCustomer.get_FullName: string;
begin
  // Put the Title, First name & Last name together
  Result := Title + ' ' + FirstName + ' ' + LastName;
end;

function TCustomer.get_LastName: string;
begin
  Result := Self.FLastName;
end;

function TCustomer.get_Title: string;
begin
  Result := Self.FTitle;
end;

procedure TCustomer.set_FirstName(Value: string);
begin
  FFirstName := Value;
end;

procedure TCustomer.set_LastName(Value: string);
begin
  FLastName := Value;
end;

procedure TCustomer.set_Title(Value: string);
begin
  FTitle := Value;
end;
{$ENDREGION}


{$REGION 'Default string representation'}
function TCustomer.ToString: string;
begin
  Result := FullName;
end;
{$ENDREGION}
{$ENDREGION}

{$REGION 'Comparers'}
{ TCompareCustomerByTitle }

function TCompareCustomerByTitle.Compare(l_customer1,
  l_customer2: TObject): integer;
begin
  Result := TCustomer( l_customer1 ).Title.CompareTo( TCustomer( l_customer2 ).Title );
end;

{ TCompareCustomerByFirstName }

function TCompareCustomerByFirstName.Compare(l_customer1,
  l_customer2: TObject): integer;
begin
  Result := TCustomer( l_customer1 ).FirstName.CompareTo( TCustomer( l_customer2 ).FirstName );
end;

{ TCompareCustomerByLastName }

function TCompareCustomerByLastName.Compare(p_customer1,
  p_customer2: TObject): integer;
begin
  if ( Assigned( p_customer1 ) and
       ( p_customer1 is TCustomer ) and
       Assigned( p_customer2 ) and
       ( p_customer2 is TCustomer ) ) then
  begin
    Result := TCustomer( p_customer1 ).LastName.CompareTo( TCustomer( p_customer2 ).LastName );
  end
  else
  begin
    raise ArgumentException.Create( 'Incorrect arguments passed to Compare' );
  end;
end;

{ TCompareCustomer }

function TCompareCustomer.Compare(p_customer1,
  p_customer2: TObject): integer;
var
  l_cust1, l_cust2: TCustomer;
  l_name1, l_name2: string;
begin
  l_cust1 := TCustomer( p_customer1 );
  l_name1 := l_cust1.LastName + l_cust1.FirstName + l_cust1.Title;
  l_cust2 := TCustomer( p_customer2 );
  l_name2 := l_cust2.LastName + l_cust2.FirstName + l_cust2.Title;
  Result := l_name1.CompareTo( l_name2 );
end;
{$ENDREGION}

{$REGION 'Utility functions'}
function TCustomer.IsChanged: boolean;
begin
  Result := ( FTitle <> FBackupData.FTitle ) or
            ( FFirstName <> FBackupData.FFirstName ) or
            ( FLastName <> FBackupData.FLastName );
end;
{$ENDREGION}

end.
